@@ -6,6 +6,6 @@
  
 #define NO_CLIP_FONT  0x7FFFFFFF
 
-extern int vgadrawtext(int,int,int,char *);
-extern int vgatextsize(int,char *);
-extern int set_max_text_width(int);
+extern int vgadrawtext(int x,int y,int siz,char *str);
+extern int vgatextsize(int sizearg,char *str);
+extern void set_max_text_width(int width);
