@@ -16,7 +16,7 @@
 
 
 
-helpscreenon()
+void helpscreenon()
 {
 vga_setmode(G640x480x16);
 vga_setcolor(15);
@@ -34,9 +34,7 @@ draw3dbox(10,10,629,469,1,0, 2,1);
 /* the caller is expected to reset the mode and redraw the screen afterwards,
  * if necessary. help is the array pointer (see vgadisp.c and zgv.c).
  */
-showhelp(ttyfd,title,help)
-int ttyfd;
-char *title,help[][80];
+void showhelp(int ttyfd,char *title,char help[][80])
 {
 int f,quit,gotkey,key;
 
