@@ -1,5 +1,5 @@
-/* Zgv v2.7 - GIF, JPEG and PBM/PGM/PPM viewer, for VGA PCs running Linux.
- * Copyright (C) 1993-1995 Russell Marks. See README for license details.
+/* Zgv v2.8 - GIF, JPEG and PBM/PGM/PPM viewer, for VGA PCs running Linux.
+ * Copyright (C) 1993-1996 Russell Marks. See README for license details.
  *
  * magic.c - Determines type of image file.
  */
@@ -9,8 +9,7 @@
 #include "magic.h"
 
 
-int magic_ident(filename)
-char *filename;
+int magic_ident(char *filename)
 {
 FILE *in;
 unsigned char buf[3];
@@ -27,10 +26,20 @@ fclose(in);
  * GIF files must have "GIF"
  * JPEG files must have first byte=0xff, second byte=0xd8 (M_SOI)
  * BMP files must start with "BM"
+ * PCX files have 0x0A, version byte, then 0x01.
  * TGA files suck rocks ;-) (heuristics in this case)
- * 
  */
 
+#ifdef NGM_SUPPORT
+/* no magic for technical reasons - test for '.ngm' extension */
+if(strlen(filename)>=4 && !strcasecmp(filename+strlen(filename)-4,".ngm"))
+  return(_IS_NGM);
+#endif
+
+/* xvpics look a bit like P?M files */
+if(!strncmp(buf,"P7 ",3))
+  return(_IS_XVPIC);
+
 /* PBM/PGM/PPM */
 if(buf[0]=='P' && buf[1]>='1' && buf[1]<='6')
   return(_IS_PNM);
@@ -53,6 +62,10 @@ if(buf[0]==0x89 && buf[1]=='P' && buf[2]=='N')
   return(_IS_PNG);		/* XXX should test the rest I s'pose */
 #endif
 
+/* PCX */
+if(buf[0]==10 && buf[2]==1)
+  return(_IS_PCX);
+
 /* TGA */
 /* this is hairy, since TGA files don't have a magic number.
  * we make a guess based on some of the image info.
