@@ -10,6 +10,9 @@
 #define _IS_BMP			4
 #define _IS_TGA			5
 #define _IS_PNG			6
-#define _IS_BAD		 	9	
+#define _IS_PCX			7
+#define _IS_NGM			8
+#define _IS_XVPIC		9
+#define _IS_BAD		 	99
 
 extern int magic_ident(char *filename);
