@@ -5,6 +5,6 @@
  */
 
 #ifdef PNG_SUPPORT
-extern int read_png_file(char *,hffunc,byte **);
-extern int aborted_file_png_cleanup();
+extern void aborted_file_png_cleanup(void);
+extern int read_png_file(char *filename, hffunc howfarfunc, byte **palette);
 #endif
