@@ -18,6 +18,6 @@ extern int read_tga_file(char *filename,hffunc howfarfunc,
 	unsigned char **bmap,unsigned char **pal,int *output_type,
 	PICINFO *pp);
 
-extern int aborted_file_tga_cleanup();
+extern void aborted_file_tga_cleanup();
 
 extern int tga_need_flip;
