@@ -1,5 +1,5 @@
-/* Zgv v2.7 - GIF, JPEG and PBM/PGM/PPM viewer, for VGA PCs running Linux.
- * Copyright (C) 1993-1995 Russell Marks. See README for license details.
+/* Zgv v2.8 - GIF, JPEG and PBM/PGM/PPM viewer, for VGA PCs running Linux.
+ * Copyright (C) 1993-1996 Russell Marks. See README for license details.
  *
  * vgadisp.c - vga specific display routines.
  */
@@ -22,6 +22,9 @@
 #include "readbmp.h"
 #include "readpng.h"
 #include "readtga.h"
+#include "readpcx.h"
+#include "readngm.h"
+#include "readxvpic.h"
 #include "readnbkey.h"
 #include "helppage.h"
 #include "3deffects.h"
@@ -42,9 +45,7 @@ int pic_incr;		/* after view, 0=stay, -1=prev, 1=next */
 char viewerhelp[][80]={
   "? (question mark) \t this help page",
   "5 to 8 \t\t mode change",
-#ifndef HANDS_ON_VC_CODE
   "F1 - F12  SF3 - SF10 \t SVGA mode change (try these first)",
-#endif
   "0 \t\t\t toggle virtual mode",
   "v \t\t\t toggle smoothing in virtual/zoom modes",
   "comma / dot \t contrast down/up (8-bit modes)",
@@ -62,6 +63,25 @@ char viewerhelp[][80]={
   ""	/* end */
   };
 
+int mode_lookup[30]={
+  /* 0-4 unused */
+  0,0,0,0,0,
+  /* 5-8 - VGA modes */
+  G320x200x256,G320x240x256,G320x400x256,G360x480x256,
+  /* 9 - unused */
+  0,
+  /* 10-29 - SVGA modes */
+  /* F1-F10 */
+  G640x480x256, G800x600x256, G1024x768x256, G1280x1024x256,
+  G320x200x32K, G320x200x64K, G320x200x16M,  G640x480x32K,
+  G640x480x64K, G640x480x16M,
+  /* SF1-SF9 */
+  G800x600x32K, G800x600x64K, G800x600x16M,  G1024x768x32K,
+  G1024x768x64K,G1024x768x16M,G1280x1024x32K,G1280x1024x64K,
+  G1280x1024x16M,
+  /* SF10 - unused */
+  0
+  };
 
 
 struct {
@@ -95,11 +115,47 @@ int saved_px,saved_py;		/* for persistance in cfg.repeat_timer mode */
 extern int zgv_ttyfd;
 int sgres;
 
-void repeat_sighandler();
-
 
+#if 1
+/* these should be quicker - an 'and' op must be quicker than a several-place
+ * bit-shift, right?
+ */
+#define GET15BITCOLOUR(r,g,b) ((((r)&0xf8)<<7)|(((g)&0xf8)<<2)|((b)>>3))
+#define GET16BITCOLOUR(r,g,b) ((((r)&0xf8)<<8)|(((g)&0xfc)<<3)|((b)>>3))
+#else
 #define GET15BITCOLOUR(r,g,b) (((r>>3)<<10)|((g>>3)<<5)|(b>>3))
 #define GET16BITCOLOUR(r,g,b) (((r>>3)<<11)|((g>>2)<<5)|(b>>3))
+#endif
+
+
+/* prototypes */
+int readpicture(char *giffn, hffunc howfarfunc, int show_dont_tell, int quick);
+int has_highcolour_mode(void);
+void fix_to_similar_mode(int modetype);
+void aborted_file_cleanup(void);
+int is_this_file_jpeg(void);
+void makerealpal(void);
+void filterpal(byte *palette);
+int dimmer(int a);
+int contrastup(int cp);
+int get_mode_width(int vm);
+int get_mode_height(int vm);
+int get_mode_pixelbytes(int vm);
+int get_mode_numcols(int vm);
+void graphicson(void);
+void graphicsoff(void);
+int showgif(byte *palette);
+void repeat_sighandler(int foo);
+void redrawgif(int px, int py, int npx, int npy);
+void eventuallydrawscansegment(byte *ptr, int x, int y, int len);
+int getvpix(int px, int py, int x, int y);
+void drawzoomedgif(void);
+void fx_mirror(void);
+void fx_flip(void);
+void fx_rot(void);
+int closest(int r, int g, int b);
+void samecentre(int *ppx, int *ppy, int newscale, int oldpx, int oldpy, int oldscale);
+void show_dimensions(int px, int py, int scaling);
 
 
 
@@ -108,11 +164,11 @@ void repeat_sighandler();
  * 1. picture is returned in global vars. 'theimage' and 'image_palette'
  * 2. you have to set pixelsize=whatever and restore it afterwards.
  * 3. you have to free() both theimage and image_palette when finished.
+ * The `quick' arg sets whether or not to ask the jpeg decoder for a
+ * rough image. This is useful when doing thumbnails, where we don't
+ * need the accuracy, but you shouldn't use it elsewhere.
  */
-int readpicture(giffn,howfarfunc,show_dont_tell)
-char *giffn;
-hffunc howfarfunc;
-int show_dont_tell;
+int readpicture(char *giffn,hffunc howfarfunc,int show_dont_tell,int quick)
 {
 int result=0;
 PICINFO ginfo;
@@ -153,7 +209,7 @@ do
         }
   
       theimage=NULL;
-      result=read_JPEG_file(giffn,howfarfunc,&palette);
+      result=read_JPEG_file(giffn,howfarfunc,&palette, quick);
       
       /* if error, then palette has already been nuked */
       if(theimage==NULL)
@@ -188,7 +244,8 @@ do
       break;
 #endif /* PNG_SUPPORT */
       
-    case _IS_PNM: case _IS_BMP: case _IS_TGA:
+    case _IS_PNM: case _IS_BMP: case _IS_TGA: case _IS_PCX:
+    case _IS_NGM: case _IS_XVPIC:
       if(show_dont_tell) pixelsize=has_highcolour_mode()?3:1;
       switch(loading_file_type)
         {
@@ -201,6 +258,15 @@ do
         case _IS_TGA:
           result=read_tga_file(giffn,howfarfunc,&theimage,&palette,&pixelsize,
                   &ginfo); break;
+	case _IS_PCX:
+          result=read_pcx_file(giffn,howfarfunc,&theimage,&palette,&pixelsize,
+                  &ginfo); break;
+	case _IS_NGM:
+          result=read_ngm_file(giffn,howfarfunc,&theimage,&palette,&pixelsize,
+                  &ginfo); break;
+	case _IS_XVPIC:
+          result=read_xvpic(giffn,howfarfunc,&theimage,&palette,&pixelsize,
+                  &ginfo); break;
         }
       
       width=ginfo.width;
@@ -273,8 +339,7 @@ return(0);
 
 
 /* since we've already tested if it's possible, this can't fail */
-fix_to_similar_mode(modetype)
-int modetype;
+void fix_to_similar_mode(int modetype)
 {
 int f,newmode,numpixels,pixdiff,newnp,newdiff,mode_ok;
 vga_modeinfo *vminfo;
@@ -333,7 +398,7 @@ if((curvgamode==7)||(curvgamode==8)) virtual=1; else virtual=0;
 }
 
 
-aborted_file_cleanup()
+void aborted_file_cleanup()
 {
 switch(loading_file_type)
   {
@@ -356,18 +421,29 @@ switch(loading_file_type)
   case _IS_PNG:
     aborted_file_png_cleanup();
     break;
-#endif /* PNG_SUPPORT */
+#endif
+#ifdef NGM_SUPPORT
+  case _IS_NGM:
+    aborted_file_ngm_cleanup();
+    break;
+#endif
+  case _IS_XVPIC:
+    fprintf(stderr,"aborted xvpic load: can't happen!\n");
+    exit(1);
+  case _IS_PCX:
+    aborted_file_pcx_cleanup();
+    break;
   }
 }
 
 
-is_this_file_jpeg()
+int is_this_file_jpeg()
 {
 return(isjpeg);
 }
 
 
-makerealpal()
+void makerealpal()
 {
 int f;
 int red=299, green=587, blue=114;	/* NTSC uses these */
@@ -383,7 +459,7 @@ for(f=0;f<256;f++)
   if(cfg.gamma)
     {
     gval=(rlo*red+glo*green+blo*blue)/4;
-    /* so gval is sub-value, 0<=gval<1000. (really <750... :( )
+    /* so gval is sub-value, 0<=gval<1000. (really <750... :-( )
      * possible emulated sub-values are...
      * 0, 114 (b), 299 (r), 413 (r+b), 587 (g), 701 (g+b), 886 (r+g)
      */
@@ -411,8 +487,7 @@ for(f=0;f<256;f++)
 }
 
 
-filterpal(palette)
-byte *palette;
+void filterpal(byte *palette)
 {
 int f;
 
@@ -427,8 +502,7 @@ makerealpal();
 }
 
 
-int dimmer(a)
-int a;
+int dimmer(int a)
 {
 a+=brightness;
 if(a<0) a=0;
@@ -437,8 +511,7 @@ return(a);
 }
 
 
-int contrastup(cp)
-int cp;
+int contrastup(int cp)
 {
 float g;
 
@@ -450,8 +523,7 @@ return((int)g);
 }
 
 
-int get_mode_width(vm)
-int vm;
+int get_mode_width(int vm)
 {
 vga_modeinfo *vmi;
 
@@ -459,8 +531,7 @@ vmi=vga_getmodeinfo(vm);
 return(vmi->width);
 }
 
-int get_mode_height(vm)
-int vm;
+int get_mode_height(int vm)
 {
 vga_modeinfo *vmi;
 
@@ -468,8 +539,7 @@ vmi=vga_getmodeinfo(vm);
 return(vmi->height);
 }
 
-int get_mode_pixelbytes(vm)
-int vm;
+int get_mode_pixelbytes(int vm)
 {
 vga_modeinfo *vmi;
 
@@ -477,8 +547,7 @@ vmi=vga_getmodeinfo(vm);
 return(vmi->bytesperpixel);
 }
 
-int get_mode_numcols(vm)
-int vm;
+int get_mode_numcols(int vm)
 {
 vga_modeinfo *vmi;
 
@@ -487,7 +556,7 @@ return(vmi->colors);
 }
 
 
-graphicson()
+void graphicson()
 {
 int x;
 
@@ -527,7 +596,8 @@ scrncols=get_mode_numcols(curvgamode);
 
 if(virtual)
   scrnwide*=2;
-vga_setpalvec(0,256,palrgb);
+if(pixelsize==1)
+  vga_setpalvec(0,256,palrgb);
 if(vkludge)
   {
   for(x=0;x<65536;x++) vkcache[x].rgb=-1;
@@ -539,7 +609,7 @@ vga_unlockvc();
 }
 
 
-graphicsoff()
+void graphicsoff()
 {
 /* we let zgv.c deal with this now, so this is just a placeholder */
 }
@@ -548,8 +618,7 @@ graphicsoff()
 /* returns 1 if ESC or 'x' was pressed, 0 if because of timeout when
  * using cfg.repeat_timer option.
  */
-int showgif(palette)   /* global theimage,height,width implied */
-byte *palette;
+int showgif(byte *palette)   /* global theimage,height,width implied */
 {
 int opx,opy,px,py,quitshow,key,redraw;
 int npx=0,npy=0;    /* equivalent to px,py in scaling>1 mode */
@@ -607,7 +676,7 @@ else
     py=saved_py;
     }
 
-redrawgif(px,py);
+redrawgif(px,py,npx,npy);
 
 if(has_mouse && mfd != -1)
   {
@@ -650,7 +719,7 @@ while(!quitshow && repeat_sig!=1)
        (((pixelsize==1)&&(get_mode_numcols(key-RK_F1+10)==256))||
         ((pixelsize==3)&&(get_mode_numcols(key-RK_F1+10)>256))))
       {
-      curvgamode=key-RK_F1+10;   /* so F1 = mode 10 */
+      curvgamode=mode_lookup[key-RK_F1+10];   /* so F1 = mode 10 */
       virtual=0;
       graphicson();
       redraw=1;
@@ -681,8 +750,14 @@ while(!quitshow && repeat_sig!=1)
   else
     switch(key)
       {
+      case 'g':
+        cfg.gamma=!cfg.gamma;
+        if(pixelsize==1)
+          filterpal(palette),vga_setpalvec(0,256,palrgb);
+        break;
       case 's': case 'd':
         if(virtual) break;
+        if(zoom) graphicson();	/* cls if zoomed, in case x2 is smaller! */
         zoom=0;
         if(scaling<512)
           {
@@ -696,8 +771,7 @@ while(!quitshow && repeat_sig!=1)
         break;
       case 'S': case 'D':
         if(virtual) break;
-        zoom=0;
-        if(scaling>1)
+        if(scaling>1 || zoom)
           {
           int oldscale=scaling;
           
@@ -707,17 +781,16 @@ while(!quitshow && repeat_sig!=1)
           redraw=1;
           if(width*scaling<scrnwide || height*scaling<scrnhigh) graphicson();
           }
+        zoom=0;
         break;
       case 'n':
         if(zoom || scaling>1)
           {
           samecentre(&px,&py,1,px,py,scaling);
           scaling=1; zoom=0;
-          redraw=1; graphicson(); 
+          graphicson(); 
           }
-        scaling=1; redraw=1;
-        if(scaling>1)  { scaling--; redraw=1; graphicson(); }
-        if(scaling==1) { px=npx; py=npy; }
+        redraw=1; interp=0; inextpix=1;
         break;
       case RK_HOME: case 'A'-0x40:
         px=py=0; break;
@@ -827,22 +900,21 @@ return(quitshow);
 }
 
 
-void repeat_sighandler()
+void repeat_sighandler(int foo)
 {
 repeat_sig=1;
 }
 
 
 /* this routine is getting ridiculous */
-redrawgif(px,py,npx,npy)
-int px,py,npx,npy;
+void redrawgif(int px,int py,int npx,int npy)
 {
 int x,y,xdim;
 int x_add,y_add;
 byte *realline,*ptr;
 
 if(zoom)
-  drawzoomedgif(height,width,theimage);
+  drawzoomedgif();
 else
   {
   int swidth=width*scaling,sheight=height*scaling;
@@ -861,11 +933,11 @@ else
     }
   
   if(swidth-px<scrnwide) xdim=swidth-px; else xdim=scrnwide;
-  if((py>=sheight)||(px>=swidth)) return(0);
+  if((py>=sheight)||(px>=swidth)) return;
   /* hopefully the following is fairly quick as fewer ifs... ? */
   if((virtual)&&(pixelsize==1))
     {
-    if((realline=calloc(1,scrnwide))==NULL) return(0);
+    if((realline=calloc(1,scrnwide))==NULL) return;
     for(y=0;(y<sheight-py)&&(y<scrnhigh);y++)
       {
       for(x=0;x<xdim;x++)
@@ -888,9 +960,9 @@ else
     unsigned char *src,*dst,*nextline;
     
     /* this doesn't allow for 'virtual' yet! */
-    if((realline=calloc(pixelsize,scrnwide))==NULL) return(0);
+    if((realline=calloc(pixelsize,scrnwide))==NULL) return;
     if(pixelsize==3)
-    if((nextline=calloc(pixelsize,scrnwide))==NULL) return(0);
+      if((nextline=calloc(pixelsize,scrnwide))==NULL) return;
     
     if(interp)
       {
@@ -995,7 +1067,7 @@ else
       
       default:
         /* for 320x200 (I think), 320x240, 320x400, 360x480 we need the
-         * length to be a multiple of 8. :(  We allocate an image-width
+         * length to be a multiple of 8. :-( We allocate an image-width
          * line (blanked), and display via that.
          */
         if(curvgamode==G320x200x256 || curvgamode==G320x240x256 ||
@@ -1003,7 +1075,7 @@ else
           {
           unsigned char *hack;
           
-          if((hack=malloc(scrnwide))==NULL) return(0);
+          if((hack=malloc(scrnwide))==NULL) return;
           for(y=0;(y<height-py)&&(y<scrnhigh);y++)
             {
             /* assume pixelsize==1, because it must be */
@@ -1020,37 +1092,35 @@ else
       }
     }
   }
-return(0);
+
+return;
 }
 
 
-eventuallydrawscansegment(ptr,x,y,len)
-byte *ptr;
-int x,y,len;
+void eventuallydrawscansegment(byte *ptr,int x,int y,int len)
 {
-unsigned short buf[2048];
+static unsigned short buf[32768];
 
 if(scrnpixelsize!=2)
   vga_drawscansegment(ptr,x,y,len);
 else
   {
-  int f;
-  unsigned short *bufref;
+  unsigned short *bufref=buf;
+  unsigned char *ptrend=ptr+len;
 
   if(scrncols==32768)
-    for(f=0,bufref=buf;f<len;f+=3,bufref++)
-      *bufref=GET15BITCOLOUR(ptr[f+2],ptr[f+1],ptr[f]);
+    for(;ptr<ptrend;ptr+=3)
+      *bufref++=GET15BITCOLOUR(ptr[2],ptr[1],*ptr);
   else
-    for(f=0,bufref=buf;f<len;f+=3,bufref++)
-      *bufref=GET16BITCOLOUR(ptr[f+2],ptr[f+1],ptr[f]);
-
-  vga_drawscansegment((byte *)buf,x,y,(len/3)<<1);
+    for(;ptr<ptrend;ptr+=3)
+      *bufref++=GET16BITCOLOUR(ptr[2],ptr[1],*ptr);
+  
+  gl_putbox(x,y,bufref-buf,1,buf);
   }
 }
 
 
-int getvpix(px,py,x,y)
-int px,py,x,y;
+int getvpix(int px,int py,int x,int y)
 {
 int p1,p2,r,g,b;
 
@@ -1070,7 +1140,7 @@ else
 
 
 /* this routine is nasty writ big, but about as quick as I can manage */
-drawzoomedgif()
+void drawzoomedgif()
 {
 register int a,b,x,yp,yw;
 long y,sw,sh,lastyp;
@@ -1081,15 +1151,15 @@ int bigimage;
 byte *rline;
 int tmp1,tmp2,tr,tg,tb,tn;
 
-if((rline=malloc(scrnwide*pixelsize))==NULL) return(0);
+if((rline=malloc(scrnwide*pixelsize))==NULL) return;
 for(x=0;x<scrnwide*pixelsize;x++)
   rline[x]=0;
 
 /* try landscapey */
-sw=scrnwide; sh=(int)((scrnwide*((long)height))/((long)width));
+sw=scrnwide; sh=(scrnwide*height)/width;
 if(sh>scrnhigh)
   /* no, oh well portraity then */
-  { sh=scrnhigh; sw=(int)((scrnhigh*((long)width))/((long)height)); }
+  sh=scrnhigh,sw=(scrnhigh*width)/height;
 
 /* so now our zoomed image will be sw x sh */
 
@@ -1123,7 +1193,7 @@ if(bigimage)
     if(yp!=lastyp)
       {
       yw=y*width;
-      if(vkludge) /* we try to resample a bit. get that pentium RSN :) */
+      if(vkludge) /* we try to resample a bit. get that pentium RSN :-) */
         {
         for(x=0;x<width;x++,yw++)
           {
@@ -1175,13 +1245,13 @@ if(bigimage)
 else
   /* well, we need to fill in the gaps because it's been made bigger. */
   {
-    /* a pixel from the original is now pixwide x pixhigh */
+  /* a pixel from the original is now pixwide x pixhigh */
   pixwide=(int) ( (((float)sw)/((float)width )) +1.0 );
   pixhigh=(int) ( (((float)sh)/((float)height)) +1.0 );
 
   free(rline);  /* get rid of that... */
     /* and get a screen's worth... hahahahahahahahaha er, sorry. */
-  if((rline=malloc(pixelsize*scrnwide*(scrnhigh+pixhigh)))==NULL) return(0);
+  if((rline=malloc(pixelsize*scrnwide*(scrnhigh+pixhigh)))==NULL) return;
     /* notice that we allow for a screwy 'pixhigh' value to avoid
        any nasty old segfaults! oh, now we wipe it: */
   for(x=0;x<scrnwide*scrnhigh*pixelsize;x++) rline[x]=0;
@@ -1221,18 +1291,17 @@ else
     				x_add,y+y_add,scrnwide*pixelsize);
   free(rline);
   }
-return(0);
 }
 
 
-fx_mirror()
+void fx_mirror()
 {
 unsigned char *src,*dst;
 byte *tmp;
 int x,y;
 
 tmp=malloc(width*pixelsize);
-if(tmp==NULL) return(0);
+if(tmp==NULL) return;
 
 if(pixelsize==1)
   {
@@ -1265,17 +1334,16 @@ else
   }
 
 free(tmp);
-return(0);
 }
 
 
-fx_flip()
+void fx_flip()
 {
 unsigned char *tmp,*ptr1,*ptr2;
 int y,y2,wp=width*pixelsize;
 
 tmp=malloc(wp);
-if(tmp==NULL) return(0);
+if(tmp==NULL) return;
 
 ptr1=theimage;
 ptr2=theimage+(height-1)*wp;
@@ -1288,11 +1356,10 @@ for(y=0,y2=height-1;y<height/2;y++,y2--,ptr1+=wp,ptr2-=wp)
   }
 
 free(tmp);
-return(0);
 }
 
 
-fx_rot()
+void fx_rot()
 {
 unsigned char *tmp,*tmp2=NULL,*ptr1,*ptr2=NULL;
 int x,y,y2,wp=width*pixelsize,hp=height*pixelsize;
@@ -1300,7 +1367,7 @@ int blockup=4;
 
 tmp=malloc(wp*height);
 if(pixelsize==3) tmp2=malloc(wp*blockup);
-if(tmp==NULL || (pixelsize==3 && tmp2==NULL)) return(0);
+if(tmp==NULL || (pixelsize==3 && tmp2==NULL)) return;
 
 if(pixelsize==1)
   for(y=0,y2=hp-1;y<height;y++,y2--)
@@ -1337,14 +1404,12 @@ if(pixelsize==3) free(tmp2);
 free(theimage);
 theimage=tmp;
 graphicson(); /* clear screen 'cos image diff. size */
-return(0);
 }
 
 
 /* rgb values must be 0..63 */
 /* because of course, only 6 bits of the rgb values actually count on VGA */
-int closest(r,g,b)
-int r,g,b;
+int closest(int r,int g,int b)
 {
 int idx,rgb;
 byte *pr,*pg,*pb;
@@ -1376,8 +1441,8 @@ return((int)(vkcache[idx].col=distnum));
 }
 
 
-samecentre(ppx,ppy,newscale,oldpx,oldpy,oldscale)
-int *ppx,*ppy,newscale,oldpx,oldpy,oldscale;
+void samecentre(int *ppx,int *ppy,int newscale,
+                int oldpx,int oldpy,int oldscale)
 {
 int xa,ya,sw,sh;
 
@@ -1408,8 +1473,7 @@ if(virtual) xa>>=1;
 /* this gets run on picture exit when cfg.selecting is true.
  * the output is designed to be put straight into a pnmcut command-line.
  */
-show_dimensions(px,py,scaling)
-int px,py,scaling;
+void show_dimensions(int px,int py,int scaling)
 {
 int x,y,w,h;
 
