@@ -16,5 +16,6 @@ extern int width,height;
 
 extern int pic_incr;	/* for ^p and ^n */
 
-extern int readpicture(char *,hffunc,int);
-extern int is_this_file_jpeg();
+extern int readpicture(char *,hffunc,int,int);
+extern int is_this_file_jpeg(void);
+extern void aborted_file_cleanup(void);
