@@ -1,5 +1,5 @@
-/* Zgv v2.7 - GIF, JPEG and PBM/PGM/PPM viewer, for VGA PCs running Linux.
- * Copyright (C) 1993-1995 Russell Marks.
+/* Zgv v2.8 - GIF, JPEG and PBM/PGM/PPM viewer, for VGA PCs running Linux.
+ * Copyright (C) 1993-1997 Russell Marks.
  *
  * zgv.c - This provides the zgv file selector, and interfaces to the
  *         vga display routines (vgadisp.c)
@@ -16,7 +16,7 @@
  * 
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
 
@@ -25,11 +25,13 @@
 #include <unistd.h>
 #include <string.h>
 #include <stdlib.h>
+#include <signal.h>
 #include <setjmp.h>
 #include <sys/param.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/file.h>
+#include <sys/ioctl.h>
 #include <sys/vt.h>
 #include <vga.h>
 
@@ -73,7 +75,7 @@ char zgvhelp[][80]={
   };
 
 #define fwinxpos(f) (40+(((f)-1)/YSIZ)*BARWIDTH)
-#define fwinypos(f) (180+barheight*(((f)-1)%YSIZ))
+#define fwinypos(f) (yofs+70+barheight*(((f)-1)%YSIZ))
 
 /* from 18-bit RGB (as used by VGA palette) to 3:3:2 palette index */
 #define MAKE332COL(r,g,b) (((r)>>3)*32+((g)>>3)*4+((b)>>4))
@@ -90,33 +92,16 @@ char zgvhelp[][80]={
 #define BAR_RESTRICT_WIDTH	BARWIDTH-6
 
 #define COLS		600
-#define MDIRSIZ		1000
-#define YSIZ		(285/barheight)
+#define MDIRSIZ		4096
+#define YSIZ		((395-yofs)/barheight)
 
 /* number of greyscales *less one* used in selector when in 640x480x16 mode */
 #define GREY_MAXVAL	10
 
-
-int idx_light,idx_medium,idx_dark,idx_black,idx_marked;
-int updating_index=0;
-int v256=1;	/* 1=256-col selector, 0=16-col selector */
-
-int barheight;
-extern int sgres;
-int has_mouse=0;
-
-
-int gcompare(void *,void *);
-void showhowfar(int,int);
-void smallhowfar(int,int);
-void exitproperly();
-void switchback();
-
-
 struct GDIR {
   char name[256];
   char isdir;            /* 0=file, 1=dir. */
-  int xvw,xvh;		/* xvpic width and height, zero if none */
+  char xvw,xvh;		/* xvpic width and height, zero if none */
   char marked;
   } gifdir[MDIRSIZ];
 int gifdirsiz;
@@ -129,17 +114,58 @@ int original_vt,separate_vt=0;
 int zgv_vt;
 int tagview_mode=0;
 
+int idx_light,idx_medium,idx_dark,idx_black,idx_marked;
+int updating_index=0;
+int v256=1;	/* 1=256-col selector, 0=16-col selector */
+
+int barheight,yofs;	/* height of cursor, and offset of selection bit */
+
+/* XXX this sgres stuff is bogus in the extreme */
+extern int sgres;
+int has_mouse=0;
 
 
-main(argc,argv)
-int argc;
-char **argv;
+/* prototypes */
+int main(int argc, char **argv);
+void openstdin_nonblocking(void);
+void load_one_file(char *filename);
+void copyfromconfig(void);
+void mainloop(void);
+int load_single_file(int curent, int do_howfar);
+void load_tagged_files(void);
+int permissiondenied(char *fname);
+void redrawall(int curent, int startfrom);
+void inithowfar(void);
+void showhowfar(int sofar, int total);
+void smallhowfar(int sofar, int total);
+void showbar(int entnum, int selected, int startfrom);
+int centreseltxt(int x, int fsiz, char *str);
+void showgifdir(int startfrom, int unshow, int drawdirmsg);
+void readgifdir(void);
+int gcompare(void *gn1, void *gn2);
+void prettyfile(char *buf, struct GDIR *gifdptr);
+void screenon(void);
+void screenoff(void);
+void drawzgvlogo(int a, int b);
+void cleartext(void);
+void showerrmessage(int errnumber);
+int delete_file(char *filename);
+void xv332_how_far(int sofar, int total);
+void clear_xvpic(int xpos, int ypos);
+int makexv332(char *filename, char *xvpicfn, unsigned int howfar);
+void update_xvpics(void);
+int fixvt(void);
+void switchback(void);
+void greyfix332(unsigned char **image, int w, int h, int xw7, int w8);
+void wait_for_foreground(void);
+void ctrlc(int foo);
+
+
+
+int main(int argc,char **argv)
 {
 int argsleft;
 
-/* kludgetastic, mate */
-if(argc>1 && strcmp(argv[1],"-d")==0) cfg.dontchangestdin=1;
-
 if(fixvt()==0)
   {
   fprintf(stderr,"Not running on console and no free VTs found.\n");
@@ -148,6 +174,9 @@ if(fixvt()==0)
 
 atexit(switchback);		/* this gets called after svgalib stuff */
 
+/* this gets called after svgalib's ^C handler, on ^C */
+signal(SIGINT,ctrlc);
+
 vga_disabledriverreport();    /* quieten svgalib */
 vga_init();
 /* root permissions should now have been ditched */
@@ -181,6 +210,8 @@ if(argsleft>1)
       gifdir[entnum].marked=1;
       gifdir[entnum].isdir=0;
       entnum++;
+      /* avoid overflowing array */
+      if(entnum>=MDIRSIZ-1) break;
       }
   gifdirsiz = entnum;
   one_file_only=1;
@@ -198,19 +229,32 @@ openstdin_nonblocking();
 mainloop();
 screenoff();
 
+if(cfg.echotagged)
+  {
+  int f,first=1;
+  
+  for(f=1;f<=gifdirsiz;f++)
+    if(gifdir[f].marked)
+      {
+      if(first) first=0; else putchar(32);
+      printf(gifdir[f].name);
+      }
+  
+  putchar('\n');
+  }
+
 exit(0);
 }
 
 
-openstdin_nonblocking()
+void openstdin_nonblocking()
 {
 zgv_ttyfd=fileno(stdin);
 fcntl(zgv_ttyfd,F_SETFL,O_NONBLOCK);
 }
 
 
-load_one_file(filename)
-char *filename;
+void load_one_file(char *filename)
 {
 hffunc hf;
 struct stat sbuf;
@@ -222,7 +266,7 @@ struct stat sbuf;
 if(stat(filename,&sbuf)!=-1 && S_ISDIR(sbuf.st_mode))
   {
   chdir(filename);
-  return(0);
+  return;
   }
   
 openstdin_nonblocking();	/* we hadn't done this yet */
@@ -241,7 +285,7 @@ else
 
 if(cfg.onefile_progress && hf!=NULL) vga_runinbackground(1);  
 
-if(readpicture(filename,hf,1)!=_PIC_OK)
+if(readpicture(filename,hf,1,0)!=_PIC_OK)
   {
   if(cfg.onefile_progress) screenoff();
   fprintf(stderr,"\rError loading file.\n");
@@ -259,7 +303,7 @@ exit(0);
 }
 
 
-copyfromconfig()
+void copyfromconfig()
 {
 curvgamode=cfg.videomode;
 zoom=cfg.zoom;
@@ -273,7 +317,7 @@ else
 }
 
 
-mainloop()
+void mainloop()
 {
 int quit,key,curent; /* quit, key pressed, current entry */
 int oldent,startfrom,oldstart,f,markchange;
@@ -285,7 +329,7 @@ quit=0; curent=1; oldent=1;
 startfrom=1;
 
 readgifdir();
-showgifdir(startfrom,0);
+showgifdir(startfrom,0,1);
 showbar(curent,1,startfrom);
 
 mx=my=1000;
@@ -349,14 +393,24 @@ while(!quit)
       
   switch(key)
     {
+    case 'f': case 'z':
+      cfg.fullsel=!cfg.fullsel;
+      redrawall(curent,startfrom);
+      break;
     case 'u':
-      if(cfg.xvpic_index)
+      /* ignore it in an xvpics dir */
+      {
+      char cdir[MAXPATHLEN+1];
+      
+      getcwd(cdir,MAXPATHLEN);
+      if(strstr(cdir,"/.xvpics")==NULL && cfg.xvpic_index)
         {
         updating_index=1;
         update_xvpics();
         updating_index=0;
         redrawall(curent,startfrom);
         }
+      }
       break;
     case 'v':
       cfg.xvpic_index=!cfg.xvpic_index;
@@ -437,11 +491,11 @@ while(!quit)
       break;
     case 'T':	/* tag all files */
       for(f=1;f<=gifdirsiz;f++) if(gifdir[f].isdir==0) gifdir[f].marked=1;
-      showgifdir(startfrom,0);
+      showgifdir(startfrom,0,0);
       break;
     case 'N':	/* untag all files */
       for(f=1;f<=gifdirsiz;f++) if(gifdir[f].isdir==0) gifdir[f].marked=0;
-      showgifdir(startfrom,0);
+      showgifdir(startfrom,0,0);
       break;
     case 'q': case 'k': case RK_CURSOR_UP:
       if(curent>1) curent--; break;
@@ -472,10 +526,10 @@ while(!quit)
         if(gifdir[curent].isdir)
           {
           showbar(curent,0,startfrom);
-          showgifdir(startfrom,1);
+          showgifdir(startfrom,1,1);
           chdir(gifdir[curent].name);
           readgifdir();
-          showgifdir(1,0); curent=oldent=startfrom=1;
+          showgifdir(1,0,1); curent=oldent=startfrom=1;
           showbar(curent,1,startfrom);
           break;
           }
@@ -524,8 +578,8 @@ while(!quit)
   if(startfrom!=oldstart)
     {
     showbar(oldent,0,oldstart);
-    showgifdir(oldstart,1);
-    showgifdir(startfrom,0);
+    showgifdir(oldstart,1,0);
+    showgifdir(startfrom,0,0);
     showbar(curent,1,startfrom);
     }
   else  
@@ -539,8 +593,7 @@ while(!quit)
 
 
 
-int load_single_file(curent,do_howfar)
-int curent,do_howfar;
+int load_single_file(int curent,int do_howfar)
 {
 static char buf[1024];
 
@@ -582,7 +635,7 @@ else
   else
     {
     if((tmp=readpicture(gifdir[curent].name,
-    		do_howfar?showhowfar:smallhowfar,1))!=_PIC_OK)
+    		do_howfar?showhowfar:smallhowfar,1, 0))!=_PIC_OK)
       {
       wait_for_foreground();
       if(tagview_mode) screenon();
@@ -596,7 +649,7 @@ return(1);
 }
 
 
-load_tagged_files()
+void load_tagged_files()
 {
 int f,t;
 int dohf;
@@ -608,7 +661,7 @@ if(t==0)
   {
   msgbox(zgv_ttyfd,"No files tagged",MSGBOXTYPE_OK,
   	idx_light,idx_dark,idx_black);
-  return(0);
+  return;
   }
 
 /* vgadisp.c sets tagview_mode==0 if esc is pressed */
@@ -623,19 +676,19 @@ do {
     if(f>1) screenon();
     msgbox(zgv_ttyfd,"Permission denied or file not found",
     	MSGBOXTYPE_OK,idx_light,idx_dark,idx_black);
-    return(0);
+    return;
     }
   if(load_single_file(f,dohf)==0) break;
   dohf=0;
   }
  } while (cfg.loop && sgres != 1);
-return(0);
+
+return;
 }
 
 
 /* this could also be file not found of course */
-int permissiondenied(fname)
-char *fname;
+int permissiondenied(char *fname)
 {
 FILE *junk;
 if((junk=fopen(fname,"rb"))==NULL)
@@ -645,16 +698,16 @@ return(0);
 }
 
 
-redrawall(curent,startfrom)
-int curent,startfrom;
+void redrawall(int curent,int startfrom)
 {
 screenon();
 
-showgifdir(startfrom,0);
+showgifdir(startfrom,0,1);
 showbar(curent,1,startfrom);
 }
 
-inithowfar()
+
+void inithowfar()
 {
 int f;
 
@@ -670,8 +723,7 @@ drawtext3d(250,235,2,"Decompressing - please wait",0,
 }
 
 
-void showhowfar(sofar,total)
-int sofar,total;
+void showhowfar(int sofar,int total)
 {
 int f,d;
 
@@ -700,8 +752,7 @@ if(((sofar%10)==0)||(sofar==total))
 
 
 /* minimal 'how far' function that just tests for an abort */
-void smallhowfar(sofar,total)
-int sofar,total;
+void smallhowfar(int sofar,int total)
 {
 /* we jump back to an abort message if Esc was pressed */
 if(!one_file_only && !cfg.repeat_timer)
@@ -717,18 +768,16 @@ if(!one_file_only && !cfg.repeat_timer)
 }
 
 
-showbar(entnum,selected,startfrom)
-int entnum,selected;
-int startfrom;
+void showbar(int entnum,int selected,int startfrom)
 {
 char ctmp[100];
 int xpos,ypos;
 int xt;
 
 xpos=fwinxpos(entnum-startfrom+1);
-if((xpos<1)||(xpos+BARWIDTH>COLS)) return(0);
+if((xpos<1)||(xpos+BARWIDTH>COLS)) return;
 ypos=fwinypos(entnum-startfrom+1);
-prettyfile(&ctmp,&(gifdir[entnum]));
+prettyfile(ctmp,&(gifdir[entnum]));
 
 set_max_text_width(BAR_RESTRICT_WIDTH);
 xt=cfg.xvpic_index?centreseltxt(xpos,GDFSIZ,ctmp):xpos+3;
@@ -777,13 +826,11 @@ else
     }
   }
 set_max_text_width(NO_CLIP_FONT);
-return(0);
+return;
 }
 
 
-int centreseltxt(x,fsiz,str)
-int x,fsiz;
-char *str;
+int centreseltxt(int x,int fsiz,char *str)
 {
 int a;
 
@@ -792,8 +839,7 @@ return(x+(BARWIDTH-a)/2);
 }
 
 
-showgifdir(startfrom,unshow)
-int startfrom,unshow;
+void showgifdir(int startfrom,int unshow,int drawdirmsg)
 {
 char cdir[MAXPATHLEN+1],*ptr;
 static char ctmp[1024],ctmp2[1024];
@@ -801,16 +847,20 @@ int f,ypos,xpos,w,h,y,xt;
 unsigned char *image;
 
 getcwd(cdir,MAXPATHLEN);
-if(updating_index)
-  sprintf(ctmp,"Updating index of %s",cdir);
-else
-  sprintf(ctmp,"Directory of %s",cdir);
 
-set_max_text_width(560);
-if(unshow)
-  undrawtext3d(40,135,4,ctmp);
-else
-  drawtext3d(40,135,4,ctmp,1, idx_light,idx_dark,idx_black);
+if(drawdirmsg)
+  {
+  if(updating_index)
+    sprintf(ctmp,"Updating index of %s",cdir);
+  else
+    sprintf(ctmp,"Directory of %s",cdir);
+  
+  set_max_text_width(560);
+  if(unshow)
+    undrawtext3d(40,yofs+25,4,ctmp);
+  else
+    drawtext3d(40,yofs+25,4,ctmp,1, idx_light,idx_dark,idx_black);
+  }
 
 set_max_text_width(BAR_RESTRICT_WIDTH);
 for(f=startfrom;f<=gifdirsiz;f++)
@@ -818,7 +868,7 @@ for(f=startfrom;f<=gifdirsiz;f++)
   xpos=fwinxpos(f-startfrom+1);
   if(xpos+BARWIDTH>COLS) break;
   ypos=fwinypos(f-startfrom+1);
-  prettyfile(&ctmp,&(gifdir[f]));
+  prettyfile(ctmp,&(gifdir[f]));
   xt=cfg.xvpic_index?centreseltxt(xpos,GDFSIZ,ctmp):xpos+3;
   vga_setcolor(unshow?idx_medium:(gifdir[f].marked?idx_marked:idx_black));
   vgadrawtext(xt,ypos+3,GDFSIZ,ctmp);
@@ -842,7 +892,13 @@ for(f=startfrom;f<=gifdirsiz;f++)
       }
     else
       {
-      sprintf(ctmp ,".xvpics/%s",gifdir[f].name);
+      /* if '/.xvpics' is in the (absolute) path somewhere, we're
+       * in a xvpics subdir. So look for the thumbnail here! :-)
+       */
+      if(strstr(cdir,"/.xvpics")!=NULL)
+        strcpy(ctmp,gifdir[f].name);
+      else
+        sprintf(ctmp ,".xvpics/%s",gifdir[f].name);
       sprintf(ctmp2,"%s/.xvpics/",getenv("HOME")?getenv("HOME"):"");
       ptr=ctmp2+strlen(ctmp2);
       getcwd(ptr,sizeof(ctmp2)-strlen(ctmp2)-1); ctmp2[strlen(ctmp2)]=0;
@@ -930,7 +986,8 @@ for(f=startfrom;f<=gifdirsiz;f++)
 set_max_text_width(NO_CLIP_FONT);
 }
 
-readgifdir()
+
+void readgifdir()
 {
 DIR *dirfile;
 struct dirent *anentry;
@@ -943,10 +1000,12 @@ dirfile=opendir(".");
 entnum=0;
 while((anentry=readdir(dirfile))!=NULL)
   {
-  if(entnum==MDIRSIZ) break;
-  if(anentry->d_name[0]=='.' && anentry->d_name[1]!='.')
-    continue;	/* skip 'dot' files */
-  if((strcmp(anentry->d_name,".")!=0)&&       /* no . and no .. if at root */
+  if(entnum>=MDIRSIZ-1) break;
+  if(!cfg.showxvpicdir || (cfg.showxvpicdir &&
+  				strcmp(anentry->d_name,".xvpics")!=0))
+    if(anentry->d_name[0]=='.' && anentry->d_name[1]!='.')
+      continue;	/* skip (most) 'dot' files */
+  if((strcmp(anentry->d_name,".")!=0) &&	/* no . and no .. if at root */
      (!((strcmp(cdir,"/")==0)&&(strcmp(anentry->d_name,"..")==0))))
     {
     if((stat(anentry->d_name,&buf))==-1)
@@ -963,6 +1022,10 @@ while((anentry=readdir(dirfile))!=NULL)
 #ifdef PNG_SUPPORT
          (!strcasecmp(anentry->d_name+l-4,".png")) ||
 #endif
+#ifdef NGM_SUPPORT
+         (!strcasecmp(anentry->d_name+l-4,".ngm")) ||
+#endif
+         (!strcasecmp(anentry->d_name+l-4,".pcx")) ||
          (!strcasecmp(anentry->d_name+l-4,".pbm")) ||
          (!strcasecmp(anentry->d_name+l-4,".pgm")) ||
          (!strcasecmp(anentry->d_name+l-4,".ppm")) || isdir)
@@ -980,8 +1043,7 @@ qsort(&(gifdir[1]),gifdirsiz,sizeof(struct GDIR),(void *)gcompare);
 }
 
 
-int gcompare(gn1,gn2)
-void *gn1,*gn2;
+int gcompare(void *gn1,void *gn2)
 {
 struct GDIR *g1,*g2;
 char buf1[80],buf2[80];
@@ -992,9 +1054,7 @@ return(strcmp(buf1,buf2));
 }
 
 
-prettyfile(buf,gifdptr)
-char *buf;
-struct GDIR *gifdptr;
+void prettyfile(char *buf,struct GDIR *gifdptr)
 {
 if(gifdptr->isdir)
   {
@@ -1007,11 +1067,13 @@ else
 }
 
 
-screenon()
+void screenon()
 {
 int r,g,b,n;
 unsigned char *tmp;
 
+yofs=(cfg.fullsel?10:110);
+
 if(cfg.xvpic_index)
   {
   if(vga_hasmode(G640x480x256) && cfg.force16fs==0)
@@ -1079,7 +1141,7 @@ if(!(cfg.xvpic_index && cfg.perfectindex))
   vga_setpalette(idx_marked,cfg.marked_r,cfg.marked_g,cfg.marked_b);
   }
 
-if(one_file_only) return(0);
+if(one_file_only) return;
 
 if(cfg.xvpic_index || v256==0)
   {
@@ -1094,25 +1156,34 @@ if(cfg.xvpic_index || v256==0)
     }
   }
 
-draw3dbox(0,0,639,99,2,1,	idx_light,idx_dark);
-draw3dbox(10,10,629,89,1,0,	idx_light,idx_dark);
-draw3dbox(0,100,639,479,2,1,	idx_light,idx_dark);
-draw3dbox(10,110,629,469,1,0,	idx_light,idx_dark);
+if(cfg.fullsel)
+  {
+  draw3dbox(0,0,639,479,2,1,	idx_light,idx_dark);
+  draw3dbox(10,10,629,469,1,0,	idx_light,idx_dark);
+  }
+else
+  {
+  draw3dbox(0,0,639,99,2,1,	idx_light,idx_dark);
+  draw3dbox(10,10,629,89,1,0,	idx_light,idx_dark);
+  draw3dbox(0,100,639,479,2,1,	idx_light,idx_dark);
+  draw3dbox(10,110,629,469,1,0,	idx_light,idx_dark);
+  }
 
-drawzgvlogo(10,10);
-return(0);
+if(!cfg.fullsel)
+  drawzgvlogo(10,10);
+
+return;
 }
 
 
-screenoff()
+void screenoff()
 {
 vga_setmode(TEXT);
 cleartext();
 }
 
 
-drawzgvlogo(a,b)
-int a,b;
+void drawzgvlogo(int a,int b)
 {
 int x,y,bw,c=0;
 byte *ptr;
@@ -1138,7 +1209,7 @@ for(y=0;y<logoh;y++)
 }
 
 
-cleartext()
+void cleartext()
 {
 if(cfg.cleartext)
   fprintf(stderr,"\033[H\033[J");
@@ -1148,12 +1219,11 @@ if(cfg.cleartext)
 /* this shows any error message from GIF reading.
  * Notice that JPEG errors have already been reported!
  */
-showerrmessage(errnumber)
-int errnumber;
+void showerrmessage(int errnumber)
 {
 char buf[256];
 
-if(updating_index) return(0);	/* ignore if updating thumbnail index */
+if(updating_index) return;	/* ignore if updating thumbnail index */
 
 strcpy(buf,"Error reading file - ");
 switch(errnumber)
@@ -1176,7 +1246,7 @@ switch(errnumber)
     strcat(buf,"can't handle RLE BMP files"); break;
   case _PICERR_SHOWN_ALREADY:
     /* this only happens with JPEGs */
-    return(0);
+    return;
 #ifdef PNG_SUPPORT
   case _PICERR_PNG_ERR:		/* a PNG error in zgv_pngerr */
     strcpy(buf,zgv_pngerr); break;
@@ -1185,7 +1255,7 @@ switch(errnumber)
     strcat(buf,"unknown error (ulp!)");
   }
 msgbox(zgv_ttyfd,buf,MSGBOXTYPE_OK, idx_light,idx_dark,idx_black);
-return(0);
+return;
 }
 
 
@@ -1201,8 +1271,7 @@ return(0);
  * OS to do the Right Thing if other thumbnail files remain
  * (which it does, of course).
  */
-delete_file(filename)
-char *filename;
+int delete_file(char *filename)
 {
 char buf[270];
 int retn=0;
@@ -1230,8 +1299,7 @@ return(retn);
 }
 
 
-void xv332_how_far(sofar,total)
-int sofar,total;
+void xv332_how_far(int sofar,int total)
 {
 char tmp[128];
 int done;
@@ -1255,8 +1323,7 @@ vga_unlockvc();
 }
 
 
-clear_xvpic(xpos,ypos)
-int xpos,ypos;
+void clear_xvpic(int xpos,int ypos)
 {
 unsigned char tmp[96];
 int y;
@@ -1270,9 +1337,7 @@ for(y=-3;y<63;y++)
 /* if howfar equals 0, no progress report is done.
  * if it is >0, then the high 16 bits are 'xpos' and low 16 are 'ypos'.
  */
-int makexv332(filename,xvpicfn,howfar)
-char *filename,*xvpicfn;
-unsigned int howfar;	/* not that signedness should really matter */
+int makexv332(char *filename,char *xvpicfn,unsigned int howfar)
 {
 FILE *out;
 int tmp;
@@ -1290,7 +1355,8 @@ if(howfar)
   xv332_how_far_ypos=howfar&0xFFFF;
   }
 
-if((tmp=readpicture(filename,howfar?xv332_how_far:NULL,0))!=_PIC_OK)
+tmp=((cfg.jpegindexstyle==3)?0:1);
+if((tmp=readpicture(filename,howfar?xv332_how_far:NULL,0,tmp))!=_PIC_OK)
   return(tmp);
 
 /* image is pointed to by theimage, image_palette */
@@ -1350,14 +1416,15 @@ return(_PIC_OK);
  * the process. Checking for Esc happens between creating xvpics.
  * the directory '.xvpics' is created if needed.
  */
-update_xvpics()
+void update_xvpics()
 {
 FILE *test;
 int f;
 int curent,oldstart,oldent,startfrom;
-struct stat realpic,xvpic;
-char buf[1024],*ptr;
+struct stat realpic,xvpic,xvpic2;
+char buf[1024],buf2[1024],*ptr;
 int redraw_needed=0;
+int r1=0,r2=0;
 
 
 /* for each picture in the current directory, we check to see if
@@ -1378,9 +1445,22 @@ for(f=1;f<=gifdirsiz;f++)
     {
     if(stat(gifdir[f].name,&realpic)==-1)
       continue;			/* the picture file doesn't exist */
-
+    
+    /* test for normal .xvpics/wibble */
     sprintf(buf,".xvpics/%s",gifdir[f].name);
-    if(stat(buf,&xvpic)==-1 || realpic.st_mtime>xvpic.st_mtime)
+    
+    /* and for ~/.xvpics/foo_bar_baz/wibble */
+    sprintf(buf2,"%s/.xvpics/",getenv("HOME")?getenv("HOME"):"");
+    ptr=buf2+strlen(buf2);
+    getcwd(ptr,sizeof(buf2)-strlen(buf2)-1); buf2[strlen(buf2)]=0;
+    while((ptr=strchr(ptr,'/'))!=NULL) *ptr++='_';
+    strcat(buf2,"/"); strcat(buf2,gifdir[f].name);
+
+    r1=stat(buf,&xvpic); r2=stat(buf2,&xvpic2);
+            
+    if((r1==-1 && r2==-1) ||
+       (r1!=-1 && realpic.st_mtime>xvpic.st_mtime) ||
+       (r2!=-1 && realpic.st_mtime>xvpic2.st_mtime))
       {
       /* this is pretty BFI and messy */
       
@@ -1402,7 +1482,7 @@ for(f=1;f<=gifdirsiz;f++)
           wait_for_foreground();
           msgbox(zgv_ttyfd,"Unable to create either .xvpics directory",
           		MSGBOXTYPE_OK,idx_light,idx_dark,idx_black);
-          return(0);
+          return;
           }
         
         /* also need to create ~/.xvpics/_foo_bar_baz */
@@ -1417,7 +1497,7 @@ for(f=1;f<=gifdirsiz;f++)
           wait_for_foreground();
           msgbox(zgv_ttyfd,"Unable to create -/.xvpics/... directory",
           		MSGBOXTYPE_OK,idx_light,idx_dark,idx_black);
-          return(0);
+          return;
           }
         
         /* make sure filename for xvpic is in buf */
@@ -1434,8 +1514,8 @@ for(f=1;f<=gifdirsiz;f++)
           screenon();
         else
           showbar(curent,0,startfrom);
-        showgifdir(startfrom,1);
-        showgifdir(startfrom,0);
+        showgifdir(startfrom,1,redraw_needed);
+        showgifdir(startfrom,0,redraw_needed);
         if(!redraw_needed)
           showbar(curent,1,startfrom);
         redraw_needed=0;
@@ -1461,8 +1541,8 @@ for(f=1;f<=gifdirsiz;f++)
     if(startfrom!=oldstart)
       {
       showbar(oldent,0,oldstart);
-      showgifdir(oldstart,1);
-      showgifdir(startfrom,0);
+      showgifdir(oldstart,1,1);
+      showgifdir(startfrom,0,1);
       showbar(curent,1,startfrom);
       }
     else  
@@ -1478,13 +1558,13 @@ for(f=1;f<=gifdirsiz;f++)
   if(readnbkey(zgv_ttyfd)==RK_ESC)
     {
     wait_for_foreground();
-    return(0);
+    return;
     }
   }
 
 wait_for_foreground();
 usleep(400000);	/* for effect :) */
-return(0);
+return;
 }
 
 
@@ -1503,7 +1583,7 @@ int num;
 char vt_filename[20];
 
 /* see if terminal is a console */
-fd=dup(fileno(stderr));
+fd=dup(2);
 fstat(fd,&sbuf);
 major=sbuf.st_rdev>>8;
 zgv_vt=minor=sbuf.st_rdev&0xff;
@@ -1518,14 +1598,15 @@ if(major==4 && minor<64)
 separate_vt=1;
 
 /* still root perms, so this shouldn't be a problem... */
-if((fd=open("/dev/console",O_RDONLY))<0) return(0);
+if((fd=open("/dev/console",O_WRONLY))<0) return(0);
 ioctl(fd,VT_GETSTATE,&vts);
 original_vt=vts.v_active;
 ioctl(fd,VT_OPENQRY,&num);
 if(num==-1) return(0);	/* no VTs free */
 
-/* now, before we go for it, we test the *current* VT to see if RW access
- * is allowed. If so, the user's probably 'genuine'.
+/* now, before we go for it, we test the *current* VT to see if they
+ * own it. If so, the user's probably `genuine'.
+ * (NB: the kernel now does this, but there's no harm repeating it.)
  */
 sprintf(vt_filename,"/dev/tty%d",original_vt);
 stat(vt_filename,&sbuf);
@@ -1535,17 +1616,29 @@ if(getuid()!=sbuf.st_uid)
   exit(1);
   }
 
-zgv_vt=num;
-sprintf(vt_filename,"/dev/tty%d",num);
-setsid();
-if(cfg.dontchangestdin==0)
-  if(freopen(vt_filename,"r",stdin)==NULL)	return(0);
-if(freopen(vt_filename,"w",stderr)==NULL)	return(0);
-/* we keep stdout - doesn't need to be changed */
-
 /* switch to the new VT */
 ioctl(fd,VT_ACTIVATE,num);
 close(fd);
+
+/* This is incredibly annoying, but the 2.0.x kernel just *will not*
+ * work without it. :-(((
+ * So, this gives really weird results for `zgv -h' etc., as the parent
+ * returns immediately. Redirect stdout if this is a problem.
+ */
+if(fork()) exit(0);
+
+zgv_vt=num;
+sprintf(vt_filename,"/dev/tty%d",num);
+
+setsid();
+
+if(freopen(vt_filename,"r",stdin)==NULL)	return(0);
+if(freopen(vt_filename,"w",stderr)==NULL)	return(0);
+/* we keep stdout - doesn't need to be changed */
+
+/* not needed, but... just in case... */
+chown(vt_filename,getuid(),getgid());
+
 /* ok, done it. */
 return(1);
 }
@@ -1572,18 +1665,7 @@ if(separate_vt)
 }
 
 
-/* prints a message saying which the old VT was if reqd. */
-oldvttell()
-{
-if(separate_vt)		/* only if needed to switch VTs */
-  fprintf(stderr,"Old VT was %d - press %sAlt-F%d to return.\n",original_vt,
-			(original_vt>12)?"Right":"",((original_vt-1)%12)+1);
-}
-
-
-greyfix332(image,w,h,xw7,w8)
-unsigned char **image;
-int w,h,xw7,w8;
+void greyfix332(unsigned char **image,int w,int h,int xw7,int w8)
 {
 unsigned char *new,*ptr;
 int x,y,lx,c;
@@ -1598,7 +1680,7 @@ if((new=malloc(w8*h))==NULL)
   {
   free(*image);
   *image=NULL;
-  return(0);
+  return;
   }
 
 memset(new,idx_medium,w8*h);
@@ -1693,12 +1775,19 @@ else
     
 free(*image);
 *image=new;
-return(0);
+return;
 }
 
 
-wait_for_foreground()
+void wait_for_foreground()
 {
 vga_runinbackground(0);
 ioctl(zgv_ttyfd,VT_WAITACTIVE,zgv_vt);
 }
+
+
+void ctrlc(int foo)
+{
+switchback();		/* and go back to blocking input */
+_exit(1);
+}
